#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <string.h>
#include <errno.h>
#include <strfunc.h>
#include <assert.h>
#include <ctype.h>

#include <common.h>
#include <header.h>

CHeadLine::CHeadLine(int F2D = 0) {
	free2destroy = F2D;
	fieldname = NULL;
	fieldbody = NULL;
	params = NULL;
	values = NULL;
	fullsize = 0;
}

CHeadLine::~CHeadLine() {
	if(free2destroy && fieldname)
		free(fieldname);

	sfree(params);
	sfree(values);
}

/*
** Fills the structure with the header line.
*/
int
CHeadLine::Set(const char *pline) {
	register char *line = (char *) pline;

	if(line == NULL || line[0] == '\0' || line[0] == ':') {
		errno = EINVAL;
		return -1;
	}

	char *colon = strchr(line, ':');
	if(colon == NULL) {
		// Neverending header.
		errno = EINVAL;
		return -1;
	} else {
		/* Skipping spaces */
		while(isspace(*++colon));
	}

	fieldname = line;
	fieldbody = colon;

	params = sinit();
	values = sinit();

	/*
	 * Splitting the parameters.
	 */
	char *param=NULL;
	char *param_end=NULL;
	char *value=NULL;
	int in_quotes=0;

	for(; *line; line++) {

		/*
		 * Escaping
		 */
		if(*line == '\\') {
			line++;	/* Eat character */
			if(!*line)
				break;
			continue;
		}

		/*
		 * Quoting
		 */
		if(*line == '"') {
			if(in_quotes) {
				in_quotes = 0;
				if(param && !param_end) {
					param_end = line;
				} else if(value) {
					sadd2(params, param, param_end - param);
					sadd2(values, value, line - value);
					param = param_end = value = NULL;
				}
			} else {
				in_quotes = 1;
				if(param && param_end && !value) {
					value = line + 1;
				} else if(!param && !param_end && !value) {
					param = line + 1;
				}
			}
			continue;
		}

		if(in_quotes)
			continue;

		if(*line == ' ') {
			if(param && !param_end) {
				param_end = line;
			} else if(value) {
				sadd2(params, param, param_end - param);
				sadd2(values, value, line - value);
				param = param_end = value = NULL;
			}
			continue;
		}

		if(*line == '=') {
			if(param && !param_end) {
				param_end = line;
			} else if(!param) {
				break;
			}
			continue;
		}

		if(*line == ';') {
			if(param && !param_end) {
				sadd(params, "");
				sadd2(values, param, line - param);
				param = value = NULL;
			} else if(value) {
				sadd2(params, param, param_end - param);
				sadd2(values, value, line - value );
				param = param_end = value = NULL;
			}
			continue;
		}

		if(*line == '\r') {
			line++;
			if(*line == '\n')
				line++;
			if(!*line)
				break;
			continue;
		}

		if(*line == '\n') {
			line++;
			if(!*line)
				break;
			continue;
		}

		if(!param) {
			param = line;
		} else if(param && param_end && !value) {
			value = line;
		}
	}


	if(param && !param_end) {
		sadd(params, "");
		sadd2(values, param, line - param);
	} else if(param && param_end && value) {
		sadd2(params, param, param_end - param);
		sadd2(values, value, line - value);
	}

	fullsize = *line ? strlen(fieldname) : (line-fieldname);
	return 0;
}


size_t
CHeadLine::size() { return fullsize; }

char *
CHeadLine::String() {
	static char empty[] = "";
	return fieldname ? fieldname : empty;
}

char *
CHeadLine::Value() {
	static char empty[] = "";
	return fieldbody ? mime_word_decode(fieldbody) : empty;
}

char *
CHeadLine::Param(char *find = NULL) {

	if(!find) {
		if(params && params->count && !*params->list[0])
			return values->list[0];
		return fieldbody;
	}

	if(!params || !params->count) {
		errno = ESRCH;
		return NULL;
	}

	for(size_t i = 1; i < params->count; i++)
		if(!strcasecmp(find, params->list[i]))
			return values->list[i];

	errno = ESRCH;
	return NULL;
}

CHeader::CHeader() {
	nHeadLines = 0;
	sHeadLines = 0;
	headLines = (CHeadLine **)NULL;
	reconstrHeader = NULL;
}

CHeader::~CHeader() {
	if(headLines)
		free(headLines);
	if(reconstrHeader)
		free(reconstrHeader);
}

/*
** Function called every time a chance to extend the headLies occur.
*/

void
CHeader::extendHeadLines() {

	// Not necessary to extend.
	if(sHeadLines && ((nHeadLines+1) < sHeadLines))
		return;

	if(sHeadLines == 0)
		sHeadLines = 8;
	else
		sHeadLines = sHeadLines << 1;

	// Resize array.
	headLines = (CHeadLine **)sf_realloc(headLines,
		sizeof(CHeadLine *) * sHeadLines);
}

int
CHeader::AddHeadLine(const char *line, int Take = 0) {

	if(!line) {
		errno = EINVAL;
		return -1;
	}

	CHeadLine *hline = new CHeadLine(Take);

	if(hline->Set(line))
		return -1;

	if(reconstrHeader) {
		/*
		 * Invalidate reconstructed header.
		 */
		free(reconstrHeader);
		reconstrHeader = NULL;
	}

	extendHeadLines();

	headLines[nHeadLines++] = hline;

	return 0;
}


char *
CHeader::HeadLineValue(char *fieldname) {

	if(!fieldname) {
		errno = EINVAL;
		return NULL;
	}

	int fieldname_len = strlen(fieldname);

	for(int i=0; i < nHeadLines; i++) {
		if(!strncasecmp(headLines[i]->String(),
				fieldname, fieldname_len))
			return headLines[i]->Value();
	}

	errno = ESRCH;
	return NULL;
}


char *
CHeader::HeadLineParam(char *fieldname, char *opt = NULL) {

	if(!fieldname) {
		errno = EINVAL;
		return NULL;
	}

	int fieldname_len = strlen(fieldname);

	for(int i=0; i < nHeadLines; i++) {
		if(!strncasecmp(headLines[i]->String(),
				fieldname, fieldname_len))
			return headLines[i]->Param(opt);
	}

	errno = ESRCH;
	return NULL;
}

size_t
CHeader::size() {
	size_t s = 1;	/* Take terminating '\n' into account */

	for(int i=0; i < nHeadLines; i++)
		s += headLines[i]->size() + 1;

	return s;
}

char *
CHeader::Save(size_t *psize = NULL) {
	size_t len = size();	/* Size of reconstructed header set */

	if(psize)
		*psize = len;

	if(reconstrHeader)
		return reconstrHeader;

	reconstrHeader = (char *)sf_malloc(len + 1);
	char *mp = reconstrHeader;

	for(int i = 0; i < nHeadLines; i++) {
		strcpy(mp, headLines[i]->String());
		mp += headLines[i]->size();
		*mp++ = '\n';
	}

	*mp++ = '\n';
	*mp = '\0';

	assert((size_t)(mp - reconstrHeader) <= len);

	return reconstrHeader;
}


